"""
TurboDesigner Airfoils Module - Contains airfoil design functionality
"""

from turbodesigner.airfoils.common import *
from turbodesigner.airfoils.DCA import *

__all__ = [
    "common",
    "DCA",
]