"""
TurboDesigner Metal Angle Methods Module - Contains methods for calculating metal angles
"""

from turbodesigner.blade.metal_angle_methods.johnsen_bullock import *

__all__ = [
    "johnsen_bullock",
]